from __future__ import annotations

import asyncio
import contextlib
import typing as tp

import aiokafka

from .base import PARAM, RT
from .base import Oxalis as _Oxalis
from .base import Task as _Task
from .base import TaskCodec, logger
from .pool import Pool

TASK_TV = tp.TypeVar("TASK_TV", bound="Task")


class Topic:
    def __init__(
        self,
        name: str,
        consumer_count: int = 3,
        pause: bool = False,
        enable_auto_commit: bool = True,
        **consumer_kwargs: tp.Any,
    ) -> None:
        self.name = name
        self.consumer_count = consumer_count
        self.pause = pause
        self.enable_auto_commit = enable_auto_commit
        self.consumer_kwargs = consumer_kwargs

    def __hash__(self) -> int:
        return hash(self.name)


class Task(_Task[PARAM, RT]):
    def __init__(
        self,
        oxalis: Oxalis,
        func: tp.Callable,
        topic: Topic,
        name: str = "",
        timeout: float = -1,
        pool: tp.Optional[Pool] = None,
    ) -> None:
        super().__init__(oxalis, func, name, timeout, pool)
        self.topic = topic
        self.key: bytes | None = None
        self.partition: int | None = None

    def config(
        self: TASK_TV,
        key: bytes | None = None,
        partition: int | None = None,
        **__,
    ) -> TASK_TV:
        self.key = key
        self.partition = partition
        return self

    def clean_config(self) -> None:
        self.key = None
        self.partition = None


class Oxalis(_Oxalis[Task]):
    def __init__(
        self,
        kafka_url: str,
        task_cls: tp.Type[Task] = Task,
        task_codec: TaskCodec = TaskCodec(),
        pool: Pool = Pool(limit=-1),
        timeout: float = 5.0,
        worker_num: int = 0,
        test: bool = False,
        group="default_group",
        default_topic=Topic("default_topic"),
        topics: tp.Sequence[Topic] = tuple(),
        producer_kwargs: tp.Dict[str, tp.Any] = {},
    ) -> None:
        super().__init__(
            task_cls,
            task_codec=task_codec,
            pool=pool,
            timeout=timeout,
            worker_num=worker_num,
            test=test,
        )
        self.kafka_url = kafka_url
        self.default_topic = default_topic
        self.group = group
        self.topics = set(topics)
        self.topics.add(self.default_topic)
        self.producer_kwargs = producer_kwargs
        self.producer: aiokafka.AIOKafkaProducer
        self.consumers: tp.List[aiokafka.AIOKafkaConsumer] = []

    async def connect(self):
        self.producer = aiokafka.AIOKafkaProducer(
            bootstrap_servers=self.kafka_url,
            request_timeout_ms=int(self.timeout * 1000),
            **self.producer_kwargs,
        )
        await self.producer.start()

    async def disconnect(self):
        await self.producer.stop()

    async def send_task(self, task: Task, *task_args, **task_kwargs):
        if task.name not in self.tasks:
            raise ValueError(f"Task {task} not register")
        logger.debug(f"Send task {task} to worker...")
        await self.producer.send_and_wait(
            task.topic.name,
            value=self.task_codec.encode(task, task_args, task_kwargs),
            key=task.key,
            partition=task.partition,
        )

    def register(
        self,
        *,
        task_name: str = "",
        timeout: float = -1,
        topic: tp.Optional[Topic] = None,
        **_,
    ) -> tp.Callable[
        [tp.Callable[PARAM, tp.Union[tp.Awaitable[RT], RT]]], Task[PARAM, RT]
    ]:
        if not topic:
            topic = self.default_topic

        def wrapped(func):
            task = self.task_cls(
                self,
                func,
                topic,
                name=task_name,
                timeout=timeout,
            )
            self.register_task(task)
            self.topics.add(topic)
            return task

        return wrapped

    async def _start_consumer(self, topic: Topic):
        self.consuming_count += 1
        consumer = aiokafka.AIOKafkaConsumer(
            topic.name,
            bootstrap_servers=self.kafka_url,
            group_id=self.group,
            enable_auto_commit=topic.enable_auto_commit,
            **topic.consumer_kwargs,
        )
        self.consumers.append(consumer)
        consumer_started = False
        try:
            await consumer.start()
            consumer_started = True
            await asyncio.sleep(
                self.timeout
            )  # wait for rebalancing when boost same topic's consumer in same time
            while self.running:
                with contextlib.suppress(asyncio.TimeoutError):
                    msg = await asyncio.wait_for(
                        consumer.getone(), timeout=self.timeout
                    )
                    if topic.pause:
                        consumer.pause(*consumer.assignment())
                    await self.on_message_receive(msg.value)
                    if topic.pause:
                        consumer.resume(*consumer.assignment())
                    if not topic.enable_auto_commit:
                        try:
                            await consumer.commit()
                        except Exception as e:
                            logger.warning(
                                f"Topic<{topic}> Commit <{consumer.assignment()}> failed: {e}"
                            )
        except Exception as e:
            self.health = False
            raise e from None
        finally:
            if consumer_started:
                await asyncio.sleep(self.timeout)  # wait for committing
                await consumer.stop()
            self.consuming_count -= 1

    def _run_worker(self):
        for t in self.topics:
            for _ in range(t.consumer_count):
                asyncio.ensure_future(self._start_consumer(t))



from __future__ import annotations

import asyncio
import typing as tp

import aio_pika

from .base import PARAM, RT
from .base import Oxalis as _Oxalis
from .base import Task as _Task
from .base import TaskCodec, logger
from .pool import Pool

TASK_TV = tp.TypeVar("TASK_TV", bound="Task")
ExchangeType = aio_pika.ExchangeType


class Exchange(aio_pika.Exchange):
    NAME_PREFIX = "oxalis_exchange_"

    def __init__(
        self,
        name: str,
        type: tp.Union[ExchangeType, str] = ExchangeType.DIRECT,
        default_routing_key: str = "",
        *,
        auto_delete: bool = False,
        durable: bool = False,
        internal: bool = False,
        passive: bool = False,
        arguments: aio_pika.abc.Arguments = None,
    ):
        self.type = self._type = type.value if isinstance(type, ExchangeType) else type
        self.name = self.NAME_PREFIX + name
        self.auto_delete = auto_delete
        self.durable = durable
        self.internal = internal
        self.passive = passive
        self.arguments = arguments or {}
        self.default_routing_key = default_routing_key

    def set_channel(self, channel: aio_pika.abc.AbstractChannel):
        self.channel = channel.channel


class Queue(aio_pika.Queue):
    NAME_PREFIX = "oxalis_queue_"

    def __init__(
        self,
        name: str,
        durable: bool = True,
        exclusive: bool = False,
        auto_delete: bool = False,
        arguments: tp.Optional[aio_pika.abc.Arguments] = None,
        passive: bool = False,
        consumer_prefetch_count: int = 4,
        consumer_prefetch_size: int = 0,
        global_: bool = False,
    ):
        self.__get_lock = asyncio.Lock()
        self.close_callbacks = aio_pika.tools.CallbackCollection(self)
        self.name = self.NAME_PREFIX + name
        self.durable = durable
        self.exclusive = exclusive
        self.auto_delete = auto_delete
        self.arguments = arguments
        self.passive = passive
        self.consumer_prefetch_count = consumer_prefetch_count
        self.consumer_prefetch_size = consumer_prefetch_size
        self.global_ = global_

    def set_channel(self, channel: aio_pika.abc.AbstractChannel):
        self.channel = channel.channel


class Task(_Task[PARAM, RT]):
    def __init__(
        self,
        oxalis: Oxalis,
        func: tp.Callable,
        exchange: Exchange,
        routing_key: str,
        name="",
        timeout: float = -1,
        ack_later: bool = True,
        ack_always: bool = False,
        reject: bool = True,
        reject_requeue: bool = False,
    ) -> None:
        super().__init__(oxalis, func, name, timeout)
        self.exchange = exchange
        self.routing_key = routing_key
        self.ack_later = ack_later
        self.ack_always = ack_always
        self.reject = reject
        self.reject_requeue = reject_requeue
        self.priority: int | None = None
        self.headers: tp.Dict[str, tp.Any] = {}
        if self.ack_always and self.reject:
            raise ValueError("'ack_always=True' conflict with 'reject=True'")
        if not self.ack_later and self.reject:
            raise ValueError("'reject=True' must get alone with 'ack_later=True'")
        if self.ack_always and not self.ack_later:
            raise ValueError("'ack_always=True' must get alone with 'ack_later=True'")
        if self.reject_requeue and not self.reject:
            raise ValueError("'reject_queue=True' need 'reject=True'")

    def config(
        self: TASK_TV, priority: int | None = None, **headers: tp.Any
    ) -> TASK_TV:
        self.priority = priority
        self.headers = headers
        return self

    def clean_config(self) -> None:
        self.priority = None
        self.headers = {}


class Oxalis(_Oxalis[Task]):
    def __init__(
        self,
        connection: aio_pika.Connection,
        task_cls: tp.Type[Task] = Task,
        task_codec: TaskCodec = TaskCodec(),
        pool: Pool = Pool(concurrency=-1),
        timeout: float = 5.0,
        worker_num: int = 0,
        test: bool = False,
        default_exchange=Exchange("default", default_routing_key="default"),
        default_queue=Queue("default"),
        default_routing_key="",
    ) -> None:
        super().__init__(
            task_cls=task_cls,
            task_codec=task_codec,
            pool=pool,
            timeout=timeout,
            worker_num=worker_num,
            test=test,
        )
        self.connection = connection
        self.default_exchange = default_exchange
        self.default_queue = default_queue
        if default_routing_key:
            self.default_exchange.default_routing_key = default_routing_key
        self.queues: tp.List[Queue] = [self.default_queue]
        self.exchanges: tp.List[Exchange] = [self.default_exchange]
        self.bindings: tp.List[tp.Tuple[Queue, Exchange, str]] = [
            (
                self.default_queue,
                self.default_exchange,
                self.default_exchange.default_routing_key,
            )
        ]
        self.routing_keys: tp.Dict[str, str] = {}
        self.channels: tp.List[aio_pika.abc.AbstractChannel] = []
        self.consumer_tags: tp.Dict[aio_pika.queue.ConsumerTag, Queue] = {}
        if self.pool.concurrency >= 0:
            raise ValueError(
                "pool concurrency config must be zero, task concurrency can be configured by Queue's Qos"
            )

    @property
    def channel(self) -> aio_pika.abc.AbstractChannel:
        if not self.channels:
            raise RuntimeError("Call connect first!")
        return self.channels[0]

    async def connect(self):
        self.connection = self.connection.__class__(
            self.connection.url, **self.connection.kwargs
        )
        await self.connection.connect(timeout=self.timeout)
        channel = self.connection.channel()
        await channel.initialize(timeout=self.timeout)
        self.channels.append(channel)
        await self.declare(self.queues)
        await self.declare(self.exchanges)
        for q, e, k in self.bindings:
            await self.bind(q, e, k)

    async def wait_close(self):
        for tag, queue in self.consumer_tags.items():
            await queue.cancel(tag, timeout=self.timeout)
        await super().wait_close()

    async def disconnect(self):
        for channel in self.channels:
            await channel.close()
        await self.connection.close()

    async def send_task(
        self,
        task: Task,
        *task_args,
        **task_kwargs,
    ):
        if task.name not in self.tasks:
            raise ValueError(f"Task {task} not register")
        task.exchange.set_channel(self.channel)
        await task.exchange.publish(
            aio_pika.Message(
                self.task_codec.encode(task, task_args, task_kwargs),
                content_type="text/plain",
                headers=task.headers,
                priority=task.priority,
            ),
            routing_key=task.routing_key,
            timeout=self.timeout,
        )

    def register(
        self,
        *,
        task_name: str = "",
        timeout: float = -1,
        exchange: tp.Optional[Exchange] = None,
        routing_key: str = "",
        ack_later: bool = True,
        ack_always: bool = False,
        reject: bool = True,
        reject_requeue: bool = False,
        **_,
    ) -> tp.Callable[
        [tp.Callable[PARAM, tp.Union[tp.Awaitable[RT], RT]]], Task[PARAM, RT]
    ]:
        if not exchange:
            exchange = self.default_exchange
        if not routing_key:
            assert exchange
            routing_key = exchange.default_routing_key

        def wrapped(func):
            task = self.task_cls(
                self,
                func,
                exchange,
                routing_key,
                name=task_name,
                timeout=timeout,
                ack_later=ack_later,
                ack_always=ack_always,
                reject=reject,
                reject_requeue=reject_requeue,
            )
            self.register_task(task)
            self.exchanges.append(task.exchange)
            return task

        return wrapped

    def on_worker_init(self):
        super().on_worker_init()
        self.channels.clear()

    def register_queues(self, queues: tp.Sequence[Queue]):
        self.queues.extend(queues)

    def register_exchanges(self, exchanges: tp.Sequence[Exchange]):
        self.exchanges.extend(exchanges)

    def register_binding(self, queue: Queue, exchange: Exchange, routing_key: str = ""):
        self.bindings.append((queue, exchange, routing_key))

    async def declare(self, eqs: tp.Sequence[tp.Union[Queue, Exchange]]):
        _names = set()
        for eq in eqs:
            if eq.name in _names:
                continue
            eq.set_channel(self.channel)
            await eq.declare(timeout=self.timeout)
            _names.add(eq.name)

    async def bind(self, queue: Queue, exchange: Exchange, routing_key: str = ""):
        queue.set_channel(self.channel)
        exchange.set_channel(self.channel)
        await queue.bind(exchange, routing_key, timeout=self.timeout)

    async def exec_task(self, task: Task, *args, **task_kwargs):
        message: aio_pika.IncomingMessage = args[0]
        task_args = args[1:]
        if not task.ack_later:
            await message.ack()
        try:
            await super().exec_task(task, *task_args, **task_kwargs)
        except Exception as e:
            if task.reject:
                await message.reject(requeue=task.reject_requeue)
            elif task.ack_always and task.ack_later:
                await message.ack()
            raise e from None
        if task.ack_later:
            await message.ack()

    async def load_and_execute_task(self, content: bytes, *args):
        try:
            message = args[0]
            task, task_args, task_kwargs = self.load_task(content)
            if self.pool.running:
                self.pool.spawn(
                    self.exec_task(task, message, *task_args, **task_kwargs),
                    timeout=task.timeout,
                )
            else:
                raise RuntimeError("Task pool closed")
        except Exception as e:
            await message.reject(requeue=True)
            logger.warning("message not consumed, so rejected it")
            raise e from None

    async def _on_message(self, message: aio_pika.abc.AbstractIncomingMessage):
        self.consuming_count += 1
        try:
            await self.on_message_receive(message.body, message)
        finally:
            self.consuming_count -= 1

    async def _receive_message(self, queue: Queue):
        channel = self.connection.channel()
        self.channels.append(channel)
        await channel.initialize()
        await channel.set_qos(
            prefetch_count=queue.consumer_prefetch_count,
            prefetch_size=queue.consumer_prefetch_size,
            global_=queue.global_,
            timeout=self.timeout,
        )
        queue.set_channel(channel)
        tag = await queue.consume(self._on_message)
        self.consumer_tags[tag] = queue

    def _run_worker(self):
        """
        Limit queue consume concurrency by AMQP's QOS config
        """
        queues = []
        _queues = set()
        for q in self.queues:
            if q.name in _queues:
                continue
            else:
                _queues.add(q.name)
                queues.append(q)

        for q in queues:
            asyncio.ensure_future(self._receive_message(q))



from __future__ import annotations

import abc
import asyncio
import inspect
import json
import logging
import multiprocessing
import os
import signal
import sys
import time
import typing as tp
from contextlib import suppress

from .pool import Pool

logger = logging.getLogger("oxalis")


TASK_TV = tp.TypeVar("TASK_TV", bound="Task")
PARAM = tp.ParamSpec("PARAM")
RT = tp.TypeVar("RT")


class Task(tp.Generic[PARAM, RT]):
    def __init__(
        self,
        oxalis: Oxalis,
        func: tp.Callable[PARAM, RT],
        name="",
        timeout: float = -1,
        pool: tp.Optional[Pool] = None,
    ) -> None:
        self.oxalis = oxalis
        self.func = func
        self.name = name or self.get_name()
        self.timeout = timeout
        self.pool = pool or oxalis.pools[0]

    def config(self: TASK_TV, **__) -> TASK_TV:
        return self

    def clean_config(self):
        pass

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.name})>"

    async def __call__(self, *args: PARAM.args, **kwargs: PARAM.kwargs) -> RT:
        ret = self.func(*args, **kwargs)
        if inspect.iscoroutine(ret):
            ret = await ret

        return ret

    async def delay(self, *args: PARAM.args, **kwargs: PARAM.kwargs) -> None:
        if self.oxalis.test:
            await self.__call__(*args, **kwargs)
        else:
            await self.oxalis.send_task(self, *args, **kwargs)
        self.clean_config()

    def get_name(self) -> str:
        return ".".join((self.func.__module__, self.func.__name__))


class TaskCodec:
    MESSAGE_TYPE = tp.Tuple[str, tp.Sequence[tp.Any], tp.Dict[str, tp.Any]]

    @classmethod
    def encode(
        cls,
        task: Task,
        task_args: tp.Sequence[tp.Any],
        task_kwargs: tp.Dict[str, tp.Any],
    ) -> bytes:
        return json.dumps([task.name, list(task_args), task_kwargs]).encode()

    @classmethod
    def decode(cls, content: bytes) -> MESSAGE_TYPE:
        return json.loads(content)


class Oxalis(abc.ABC, tp.Generic[TASK_TV]):
    READY_FILE_PATH: tp.ClassVar[str] = "/tmp/oxalis_ready"
    HEARTBEAT_FILE_PATH: tp.ClassVar[str] = "/tmp/oxalis_heartbeat"

    def __init__(
        self,
        task_cls: tp.Type[TASK_TV],
        task_codec: TaskCodec = TaskCodec(),
        pool: Pool = Pool(),
        timeout: float = 5.0,
        worker_num: int = 0,
        test: bool = False,
    ) -> None:
        self.task_cls = task_cls
        self.tasks: tp.Dict[str, TASK_TV] = {}
        self.task_codec = task_codec
        self.pools: tp.List[Pool] = [pool]
        self.running = False
        self.timeout = timeout
        self.test = test
        self._on_close_signal_count = 0
        self.worker_num = worker_num or os.cpu_count()
        self.is_worker = False
        self.consuming_count = 0
        self.health = True

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(pid-{os.getpid()})>"

    @property
    def pool(self) -> Pool:
        return self.pools[0]

    async def connect(self):
        pass

    async def wait_close(self):
        while self.consuming_count:
            await asyncio.sleep(self.timeout)

    async def disconnect(self):
        pass

    @abc.abstractmethod
    async def send_task(
        self,
        task: TASK_TV,
        *task_args,
        **task_kwargs,
    ):
        pass

    async def exec_task(self, task: TASK_TV, *task_args, **task_kwargs):
        logger.debug(f"Worker {self} execute task {task}...")
        await task(*task_args, **task_kwargs)

    def run_worker_master(self):
        for task in self.tasks.values():
            logger.info(f"Registered Task: {task}")

        signal.signal(signal.SIGINT, self.close)
        signal.signal(signal.SIGTERM, self.close)
        ps = []
        for _ in range(self.worker_num):
            ps.append(multiprocessing.Process(target=self.run_worker))
            ps[-1].start()
        for p in ps:
            p.join()

    def run_worker(self):
        logger.info(f"Run worker: {self}...")
        self.running = True
        self.is_worker = True
        self.on_worker_init()
        asyncio.get_event_loop().run_until_complete(self.connect())
        self._run_worker()
        asyncio.get_event_loop().run_until_complete(self.work())
        self.on_worker_close()

    @abc.abstractmethod
    def _run_worker(self):
        pass

    async def work(self):
        with open(self.READY_FILE_PATH, "w") as f:
            f.write(f"{time.time():.0f}\n")
        while self.running:
            if self.health:
                with open(self.HEARTBEAT_FILE_PATH, "w") as f:
                    f.write(f"{time.time():.0f}\n")
            await asyncio.sleep(self.timeout)

        await self.wait_close()
        await asyncio.wait(
            [asyncio.get_event_loop().create_task(p.wait_close()) for p in self.pools],
        )
        await self.disconnect()
        with suppress(FileNotFoundError):
            os.remove(self.READY_FILE_PATH)
        with suppress(FileNotFoundError):
            os.remove(self.HEARTBEAT_FILE_PATH)

    def close_worker(self, force: bool = False):
        logger.info(f"Close worker{'(force)' if force else ''}: {self}...")
        self.running = False
        if force:
            logger.warning(f"Force close: {self}, may lose some message!")
            for p in self.pools:
                p.force_close()
            sys.exit()

    def register_task(self, task: TASK_TV):
        if task.name in self.tasks:
            raise ValueError("double task, check task name")
        self.tasks[task.name] = task

    def register(
        self,
        *,
        task_name: str = "",
        timeout: float = -1,
        pool: tp.Optional[Pool] = None,
        **_,
    ) -> tp.Callable[
        [tp.Callable[PARAM, tp.Union[tp.Awaitable[RT], RT]]], Task[PARAM, RT]
    ]:
        def wrapped(func: tp.Callable[PARAM, RT]):
            task = self.task_cls(self, func, name=task_name, timeout=timeout, pool=pool)
            self.register_task(task)
            return task

        return wrapped

    def load_task(self, content: bytes) -> tp.Tuple[TASK_TV, tp.Sequence, tp.Dict]:
        task_name, task_args, task_kwargs = self.task_codec.decode(content)
        if task_name not in self.tasks:
            raise ValueError(f"task_name {task_name} not founded!")

        return self.tasks[task_name], task_args, task_kwargs

    async def load_and_execute_task(self, content: bytes, *_):
        task, task_args, task_kwargs = self.load_task(content)
        await asyncio.wait_for(
            self.exec_task(task, *_, *task_args, **task_kwargs),
            timeout=self.pool.timeout if task.timeout == -1 else task.timeout,
        )

    async def on_message_receive(self, content: bytes, *_):
        """should not raise exception"""
        try:
            return await self.load_and_execute_task(content, *_)
        except Exception as e:
            logger.exception(e)

    def close(self, *_):
        """Close self but wait pool"""
        if not self.is_worker:
            return
        self._on_close_signal_count += 1
        self.close_worker(force=self._on_close_signal_count >= 2)

    def on_worker_init(self):
        pass

    def on_worker_close(self):
        pass



from __future__ import annotations

import asyncio
import time
import typing as tp
import uuid

from redis.asyncio.client import Redis

from .base import PARAM, RT
from .base import Oxalis as _Oxalis
from .base import Task as _Task
from .base import TaskCodec, logger
from .pool import Pool

TASK_TV = tp.TypeVar("TASK_TV", bound="Task")


class Queue:
    NAME_PREFIX = "oxalis_queue_"

    def __init__(self, name: str, consumer_count: int = 3) -> None:
        self.name = self.NAME_PREFIX + name
        self.consumer_count = consumer_count

    def __hash__(self) -> int:
        return hash(self.name)


class PubsubQueue(Queue):
    def __init__(self, name: str, consumer_count: int = 1) -> None:
        super().__init__(name, consumer_count=consumer_count)
        if self.consumer_count != 1:
            raise ValueError("Wrong consumer count for pubsub queue")


class Task(_Task[PARAM, RT]):
    def __init__(
        self,
        oxalis: Oxalis,
        func: tp.Callable,
        queue: Queue,
        name="",
        timeout: float = -1,
        pool: tp.Optional[Pool] = None,
    ) -> None:
        super().__init__(oxalis, func, name, timeout, pool)
        self.queue = queue
        self.delay_timeout: int | None = None

    def config(
        self: TASK_TV,
        delay_timeout: int | None = None,
        **__,
    ) -> TASK_TV:
        self.delay_timeout = delay_timeout
        return self

    def clean_config(self) -> None:
        self.delay_timeout = None


class Oxalis(_Oxalis[Task]):
    def __init__(
        self,
        client: Redis,
        task_cls: tp.Type[Task] = Task,
        task_codec: TaskCodec = TaskCodec(),
        pool: Pool = Pool(),
        timeout: float = 2.0,
        worker_num: int = 0,
        test: bool = False,
        default_queue_name: str = "default",
        delay_queue_name: str = "delay_default",
    ) -> None:
        super().__init__(
            task_cls=task_cls,
            task_codec=task_codec,
            pool=pool,
            timeout=timeout,
            worker_num=worker_num,
            test=test,
        )
        self.client = client
        self.pubsub = client.pubsub()
        self.default_queue = Queue(default_queue_name)
        self.delay_queue = Queue(delay_queue_name)
        self._receiving = False

    async def connect(self):
        await self.client.initialize()

    async def wait_close(self):
        while self.pubsub.connection is not None:
            await asyncio.sleep(self.timeout)
        await super().wait_close()

    async def disconnect(self):
        await self.client.close()

    async def send_task(self, task: Task, *task_args, **task_kwargs):
        if task.name not in self.tasks:
            raise ValueError(f"Task {task} not register")
        content = self.task_codec.encode(task, task_args, task_kwargs)
        if task.delay_timeout:
            logger.debug(
                f"Send task {task} to worker with {task.delay_timeout}s delay..."
            )
            await self.client.zadd(
                self.delay_queue.name,
                {uuid.uuid1().bytes + content: time.time() + task.delay_timeout},
            )
        else:
            logger.debug(f"Send task {task} to worker...")
            if isinstance(task.queue, PubsubQueue):
                await self.client.publish(task.queue.name, content)
            else:
                await self.client.rpush(task.queue.name, content)

    def register(
        self,
        *,
        task_name: str = "",
        timeout: float = -1,
        queue: tp.Optional[Queue] = None,
        **_,
    ) -> tp.Callable[
        [tp.Callable[PARAM, tp.Union[tp.Awaitable[RT], RT]]], Task[PARAM, RT]
    ]:
        def wrapped(func):
            task = self.task_cls(
                self,
                func,
                queue or self.default_queue,
                name=task_name,
                timeout=timeout,
            )
            self.register_task(task)
            return task

        return wrapped

    async def _schedule_delayed_message(
        self, fetch_count: int = 100, time_offset: float = 0.05
    ):
        while self.running:
            now_ts = time.time()
            count = 0
            for content, _ in await self.client.zrangebyscore(
                self.delay_queue.name,
                min=0,
                max=now_ts + time_offset,
                start=0,
                num=fetch_count,
                withscores=True,
            ):
                count += 1
                if await self.client.zrem(
                    self.delay_queue.name, content
                ):  # avoid double schedule
                    try:
                        content = content[16:]
                        task_name, _, _ = self.task_codec.decode(content)
                        if task_name not in self.tasks:
                            logger.warning(f"Received task {task_name} not found")
                        else:
                            task = self.tasks[task_name]
                            if isinstance(task.queue, PubsubQueue):
                                await self.client.publish(task.queue.name, content)
                            else:
                                await self.client.rpush(task.queue.name, content)
                    except Exception as e:
                        logger.exception(e)
            if count < fetch_count:
                await asyncio.sleep(time_offset)

    async def _receive_message(self, queue: Queue):
        self.consuming_count += 1
        try:
            while self.running:
                content = await self.client.blpop(queue.name, timeout=self.timeout)
                if not content:
                    continue
                else:
                    await self.on_message_receive(content[1])
        except Exception as e:
            self.health = False
            raise e from None
        finally:
            self.consuming_count -= 1

    async def _receive_pubsub_message(self, queue: Queue):
        self.consuming_count += 1
        try:
            while self.running:
                if not self.pubsub.subscribed:
                    await self.pubsub.subscribe(queue.name)
                content = await self.pubsub.get_message(
                    ignore_subscribe_messages=True, timeout=self.timeout
                )
                if not content:
                    continue
                else:
                    await self.on_message_receive(content["data"])
        except Exception as e:
            self.health = False
            raise e from None
        finally:
            self.consuming_count -= 1
            await self.pubsub.close()

    def _run_worker(self):
        """
        Limit queue consume concurrency by pool
        """
        queue_names = set()
        pubsub_queue_names = set()
        for t in self.tasks.values():
            if isinstance(t.queue, PubsubQueue):
                pubsub_queue_names.add(t.queue)
            else:
                queue_names.add(t.queue)
        for queue in queue_names:
            for _ in range(queue.consumer_count):
                asyncio.ensure_future(self._receive_message(queue))
        for queue in pubsub_queue_names:
            for _ in range(queue.consumer_count):
                asyncio.ensure_future(self._receive_pubsub_message(queue))
        asyncio.ensure_future(self._schedule_delayed_message())



import asyncio
import logging
import typing as tp
from asyncio.queues import Queue, QueueEmpty

logger = logging.getLogger("oxalis_pool")


class Pool:
    def __init__(
        self,
        limit: int = 0,
        concurrency: int = 100,
        timeout: tp.Union[int, float, None] = 5 * 60,
    ):
        self.concurrency = limit or concurrency
        self.timeout = timeout
        self.pending_queue: Queue[tp.Tuple[tp.Awaitable, float | None]] = Queue()
        self.done_queue: Queue = Queue()
        self.futures: tp.Set[asyncio.Future] = set()
        self.running_count = 0
        self.running = True

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} <running_count: {self.running_count}, pending_count: {self.pending_queue.qsize()}>"

    def spawn(
        self, coroutine: tp.Awaitable, pending: bool = True, timeout: float | None = -1
    ) -> tp.Optional[asyncio.Future]:
        if not self.running:
            raise RuntimeError("This pool has been closed")

        if self.concurrency < 0 or self.running_count < self.concurrency:
            return self._ensure_future(coroutine, timeout=timeout)
        elif pending:
            self.pending_queue.put_nowait((coroutine, timeout))
            return None
        else:
            return None

    def ensure_future(
        self, coroutine: tp.Awaitable, pending: bool = True, timeout: float | None = -1
    ) -> tp.Optional[asyncio.Future]:
        return self.spawn(coroutine, pending=pending, timeout=timeout)

    def _ensure_future(self, coroutine: tp.Awaitable, timeout: float | None):
        timeout = self.timeout if timeout == -1 else timeout
        self.running_count += 1
        f = asyncio.ensure_future(asyncio.wait_for(coroutine, timeout=timeout))
        f.add_done_callback(self.on_future_done)
        self.futures.add(f)
        return f

    async def wait_spawn(
        self, coroutine: tp.Awaitable, timeout: float = -1
    ) -> tp.Optional[asyncio.Future]:
        while self.running:
            f = self.spawn(coroutine, pending=False, timeout=timeout)
            if f:
                return f
            else:
                await self.done_queue.get()
        return None

    @property
    def done(self) -> bool:
        return not (self.running_count or self.pending_queue.qsize())

    async def wait_done(self):
        while not self.done:
            await self.done_queue.get()

    def close(self, force=False):
        logger.info(f"Close {'(force)' if force else ''} {self}...")
        self.running = False
        if force:
            self.running = False
            while not self.pending_queue.empty():
                self.pending_queue.get_nowait()
            for f in self.futures:
                f.cancel()

    async def wait_close(self):
        self.close()
        await self.wait_done()

    def force_close(self):
        self.close(force=True)

    def check_future(self, f: asyncio.Future):
        e = f.exception()
        if e:
            try:
                raise e
            except Exception:
                logger.exception(e)

    def on_future_done(self, f: asyncio.Future):
        self.running_count -= 1
        self.done_queue.put_nowait(f)
        self.check_future(f)
        self.futures.remove(f)
        try:
            if self.concurrency == -1 or self.running_count < self.concurrency:
                next, timeout = self.pending_queue.get_nowait()
                self._ensure_future(next, timeout)
        except QueueEmpty:
            pass
